#include <exception.h>
#include <ext2.h>
#include <macros.h>
#include <utils.h>

// TODO: Este include no debería ser siempre necesario
#include <ext2_write.h>

/*
 * Los tipos que aparecen en las entradas de directorio
 */
static const uint8_t ext2_file_types[8] = {
    FILE_TYPE_NULL,        FILE_TYPE_REGULAR_FILE, FILE_TYPE_DIRECTORY,
    FILE_TYPE_CHAR_DEVICE, FILE_TYPE_BLOCK_DEVICE, FILE_TYPE_FIFO,
    FILE_TYPE_UNIX_SOCKET, FILE_TYPE_SYMBOLIC_LINK};

/**
 * Devuelve el tamaño del archivo correspondiente al inodo
 */
uint64_t ext2_get_size(ext2_inode* inode) {
    uint64_t size = 0;
    // Si no es un directorio entonces la parte alta cuenta
    // Si es un directorio entonces es parte del ACL
    if (!IS_MODE_TYPE(inode->mode, EXT2_INODE_TYPE_DIRECTORY))
        size = (uint64_t)inode->size_high << 32;
    return size | inode->size_low;
}
/**
 * Parsea el superbloque y guarda informacion del filesystem en info
 * @returns Devuelve el numero de bloque donde se encuentra el superbloque
 */
uint32_t ext2_parse_superblock(fs_info* info) {
    block_cache* sb_cache = block_cache_create(EXT2_SUPERBLOCK_SIZE, 1);
    ext2_superblock* sb = block_cache_load(sb_cache, EXT2_SUPERBLOCK_INDEX);
    // Chequeamos las features
    ext2_check_superblock(sb);
    // Cargamos algunos datos adicionales
    info->block_shift = EXT2_MIN_BLOCK_SHIFT + sb->block_size_shift;
    info->block_size = 1 << info->block_shift;
    // Division entera con techo para la cantidad de grupos
    info->group_count = sb->number_blocks / sb->blocks_per_group +
                        (sb->number_blocks % sb->blocks_per_group != 0);
    uint32_t bn_superblock = sb->bn_superblock;
    // Esta cache era temporal para leer info del superbloque
    block_cache_free(sb_cache);
    return bn_superblock;
}

void ext2_check_superblock(ext2_superblock* sb) {
    // Chequeamos la magia
    if (sb->ext2_magic != EXT2_MAGIC) {
        KERNEL_PANIC("No se ha detectado un file system, magia invalida");
    }
    // Que no tenga errores
    if (sb->fs_state != EXT2_STATE_CLEAN) {
        KERNEL_PANIC("El file system tiene errores");
    }
    // Que no tenga mas features que las que podemos soportar
    if ((sb->req_flags & ~REQUIRED_FLAGS_SUPPORT) > 0) {
        KERNEL_PANIC("El file system tiene parametros que no son soportados");
    }
}

/**
 * Carga un inodo, usa la cache de bloques
 * Hay que deallocarlo despues con block_cache_unload
 * @returns Puntero al inodo en la cache del bloque donde estaba el inodo
 */
ext2_inode* ext2_load_inode(fs_info* info, uint32_t inode_number) {
    uint32_t inodes_per_group = info->sb->inodes_per_group;
    uint32_t inode_size = info->sb->inode_size;
    uint32_t block_size = info->block_size;
    // A que grupo pertence
    uint32_t group = (inode_number - 1) / inodes_per_group;
    // Indice en la tabla del grupo
    uint32_t index = (inode_number - 1) % inodes_per_group;
    // Block number absoluto donde se encuentra el inodo
    uint32_t block_number =
        info->gt[group].bn_inode_table + (index * inode_size) / block_size;
    // Su indice dentro del bloque como un array de inodos
    uint32_t offset = index % (block_size / inode_size);
    // Este bloque despues se descarga con block_cache_unload
    return ((ext2_inode*)block_cache_load(info->cache, block_number)) + offset;
}

uint64_t ext2_read(block_cache* cache, uint64_t block_shift, fs_file* file,
                   void* buffer, uint64_t count) {
    // Calculamos el tamaño del bloque
    uint64_t block_size = 1 << block_shift;
    uint64_t total = MIN(count, file->remaining);
    uint64_t left = total;
    // Ciclo que se encarga de copiar los bytes
    // del elemento sin importar cuantos bloques ocupe.
    while (left) {
        uint64_t to_copy = MIN(left, block_size - file->block_offset);
        void* block_ptr = block_cache_load(cache, file->block_number);
        void* base = (void*)((uintptr_t)block_ptr + file->block_offset);
        memcpy(buffer, base, to_copy);
        block_cache_unload(cache, block_ptr);
        ext2_seek(cache, block_shift, file, to_copy);
        buffer = (void*)((uintptr_t)buffer + to_copy);
        left -= to_copy;
    }
    return total;
}

/**
 * @param offset El offset en bytes, puede ser negativo,
 * Si se da uno invalido la funcion devuelve -1
 */
int64_t ext2_seek(block_cache* cache, uint64_t block_shift, fs_file* file,
                  int64_t offset) {
    // Calculamos el tamaño de bloque
    uint64_t block_size = 1 << block_shift;
    // Upper bound check
    if (offset > (int64_t)file->remaining)
        return -1;
    // WARNING: Shifting de negativos a la derecha es implementation defined
    // Calculamos que tantos bloques hay que avanzar, con negativos es flooring
    int64_t advance = (offset + (int64_t)file->block_offset) >> block_shift;
    int64_t index = advance + file->block_index;
    // Lower bound check
    if (index < 0)
        return -1;
    // Ajustamos el offset dentro del bloque
    file->block_offset += offset;
    file->block_offset &= block_size - 1;
    // Ajustamos los bytes restantes del archivo
    file->remaining -= offset;
    // Si no nos movimos del bloque no hay que hacer nada mas
    if (!advance)
        return 0;
    // Si nos movimos ajustamos el nuevo indice
    file->block_index = index;
    // Si justo estamos en el final del ultimo bloque
    // No tenemos nada más que cargar más
    if (!file->remaining)
        return 0;
    // Si es directo, no tenemos que hacer nada mas
    if (index < 12) {
        file->block_number = file->inode->bn_direct[index];
        return 0;
    }
    // En un bloque entran (block_size / 4) numeros de bloque
    uint64_t bn_shift = block_shift - 2;
    // El indice relativo del bloque indirecto simple
    uint64_t new_single = ((index - 12) >> bn_shift) + 1;
    // Si es mas allá del bloque indirecto simple explicito
    if (new_single > 1) {
        // El indice relativo del bloque indirecto doble
        uint64_t new_double = ((new_single - 2) >> bn_shift) + 1;
        // Si necesitamos cargar el triple indirecto
        if (new_double > 1 && file->triply_indirect == NULL)
            file->triply_indirect =
                block_cache_load(cache, file->inode->bn_triple);
        // Si necesitamos cargar el doble indirecto
        if (file->double_index != new_double) {
            // Actualizamos el valor y lo descargamos
            file->double_index = new_double;
            block_cache_unload(cache, file->doubly_indirect);
            // Bloque indirecto doble a cargar
            uint32_t double_block;
            if (new_double == 1)
                double_block = file->inode->bn_double;
            else
                double_block = file->triply_indirect[new_double - 2];
            file->doubly_indirect = block_cache_load(cache, double_block);
        }
    }
    // Mascara para sacar el resto de dividir por cantidad de numeros de bloque
    uint64_t bn_mask = (1 << bn_shift) - 1;
    // Si necesitamos cargar el simple indirecto
    // new_single nunca va a ser 0 porque estamos arriba de los 12 directos
    if (file->single_index != new_single) {
        // Actualizamos el valor y lo descargamos
        file->single_index = new_single;
        block_cache_unload(cache, file->singly_indirect);
        // Bloque indirecto simple a cargar
        uint32_t single_block;
        if (new_single == 1)
            single_block = file->inode->bn_single;
        else
            single_block = file->doubly_indirect[(new_single - 2) & bn_mask];
        file->singly_indirect = block_cache_load(cache, single_block);
    }
    file->block_number = file->singly_indirect[(index - 12) & bn_mask];
    return 0;
}

int64_t fs_mount(fs_info* info) {
    // Parseamos data inicial
    uint32_t bn_superblock = ext2_parse_superblock(info);
    uint64_t block_size = info->block_size;
    // Creamos la cache de bloques que vamos a usar
    info->cache = block_cache_create(block_size, EXT2_INITIAL_CACHE_COUNT);
    // Medio dirty pero es mejor que lo que teniamos antes
    // Creamos una cache con tamaños de bloque hasta la tabla
    // Y cargamos un solo bloque
    uint64_t group_start = (bn_superblock + 1) * block_size;
    uint64_t head_size =
        group_start + info->group_count * sizeof(ext2_group_desc);
    info->head_cache = block_cache_create(head_size, 1);
    uintptr_t head_ptr = (uintptr_t)block_cache_load(info->head_cache, 0);
    info->sb = (ext2_superblock*)(head_ptr + EXT2_SUPERBLOCK_OFFSET);
    info->gt = (ext2_group_desc*)(head_ptr + group_start);
    return 0;
}

void fs_unmount(fs_info* info) {
    block_cache_free(info->head_cache);
    block_cache_free(info->cache);
    // TODO: Esto es de ext2_write
    if (info->sb->req_write_flags & EXT2_FEATURE_SPARSE)
        ext2_restore_backup(info);
}

/*
 * Descarga los bloques indirectos por ahora
 */
void ext2_unload_file(fs_info* info, fs_file* file) {
    if (!file)
        return;
    // TODO: Esto es de ext2_write
    if (file->prealloc_count > 0)
        ext2_set_resvd_blocks(info, file->prealloc_block, file->prealloc_count,
                              false);

    block_cache* cache = info->cache;

    block_cache_unload(cache, file->singly_indirect);
    block_cache_unload(cache, file->doubly_indirect);
    block_cache_unload(cache, file->triply_indirect);
}

/*
 * Carga file con los datos iniciales de un archivo
 */
fs_file* ext2_load_file(ext2_inode* inode, fs_file* file, uint32_t inode_num) {
    file->inode = inode;
    file->remaining = ext2_get_size(inode);
    file->block_index = 0;
    file->block_offset = 0;
    file->block_number = inode->bn_direct[0];
    file->single_index = 0;
    file->double_index = 0;
    file->singly_indirect = NULL;
    file->doubly_indirect = NULL;
    file->triply_indirect = NULL;

    file->prealloc_count = 0;
    file->prealloc_block = 0;
    file->inode_number = inode_num;
    return file;
}

fs_dir* ext2_load_dir(ext2_inode* inode, fs_dir* dir, uint32_t inode_num) {
    ext2_load_file(inode, &dir->file, inode_num);
    dir->rec_inode = 0;
    return dir;
}

/**
 * Devuelve la posición absoluta en el archivo file.
 */
int64_t fs_tell(fs_info* info, fs_file* file) {
    return (file->block_index << info->block_shift) + file->block_offset;
}

int64_t fs_fstat(fs_info* info, fs_file* file, fs_stat* stat) {
    stat->size = ext2_get_size(file->inode);
    stat->mode = file->inode->mode;
    return 0;
}

fs_dir* fs_root(fs_info* info, fs_dir* dir) {
    return ext2_load_dir(ext2_load_inode(info, EXT2_ROOT_INODE), dir,
                         EXT2_ROOT_INODE);
}

fs_file* fs_open(fs_info* info, fs_dir* dir, fs_file* dest) {
    if (!dir->rec_inode)
        return NULL;
    uint32_t dir_inode = dir->rec_inode;
    return ext2_load_file(ext2_load_inode(info, dir_inode), dest, dir_inode);
}

fs_dir* fs_opendir(fs_info* info, fs_dir* dir, fs_dir* dest) {
    uint32_t inode = dir->rec_inode;
    int8_t file_type = dir->data.file_type;
    if (dir == dest)
        fs_closedir(info, dest);
    if (!inode || file_type != FILE_TYPE_DIRECTORY)
        return NULL;
    return ext2_load_dir(ext2_load_inode(info, inode), dest, inode);
}

/*
 * Cierra un archivo para poder ser liberado
 */
int64_t fs_close(fs_info* info, fs_file* file) {
    block_cache_unload(info->cache, file->inode);
    ext2_unload_file(info, file);
    // En el futuro se podría devolver algún error si fallara algo.
    return 0;
}

int64_t fs_closedir(fs_info* info, fs_dir* dir) {
    fs_close(info, &dir->file);
    return 0;
}

fs_dir* fs_rewinddir(fs_info* info, fs_dir* dir) {
    int64_t back = -fs_tell(info, &dir->file);
    ext2_seek(info->cache, info->block_shift, &dir->file, back);
    dir->rec_inode = 0;
    return dir;
}

/*
 * Lee el proximo directorio de dir y lo pone en dir->data
 * @returns NULL si no hay proximo, o el puntero a dir->data
 */
dir_entry* fs_readdir(fs_info* info, fs_dir* dir) {
    uint64_t shift = info->block_shift, next_entry = 0;
    fs_file* file = &dir->file;
    block_cache* cache = info->cache;
    ext2_dir_entry entry;
    entry.inode_number = 0;
    // Buscamos la primera entry no nula
    while (ext2_read(cache, shift, file, &entry, sizeof(ext2_dir_entry))) {
        next_entry = entry.rec_len - sizeof(ext2_dir_entry);
        if (entry.inode_number)
            break;
        ext2_seek(cache, shift, file, next_entry);
    }
    // Si llegamos al final y no encontramos ninguna retornamos null
    if (!entry.inode_number)
        return NULL;
    // Guardamos el tipo de archivo e inodo
    dir->rec_inode = entry.inode_number;
    dir->data.file_type = ext2_file_types[entry.file_type];
    // El puntero de bloque queda justo en el nombre del archivo.
    // Entonces solo basta hacer un kfread.
    ext2_read(cache, shift, file, dir->data.name, entry.name_len);
    // Ext2 no asegura null terminated strings
    dir->data.name[entry.name_len] = '\0';
    dir->rec_len = entry.rec_len;
    // Para avanzar hasta la próxima entrada tengo que avanzar lo que me queda
    ext2_seek(cache, shift, file, next_entry - entry.name_len);
    return &dir->data;
}

uint64_t fs_read(fs_info* info, fs_file* file, void* buffer, uint64_t count) {
    return ext2_read(info->cache, info->block_shift, file, buffer, count);
}

int64_t fs_seek(fs_info* info, fs_file* file, int64_t offset) {
    return ext2_seek(info->cache, info->block_shift, file, offset);
}
