#include <bitmap.h>
#include <exception.h>
#include <ext2_write.h>
#include <macros.h>
#include <utils.h>

#define MULTIPLE_OF_4(x) ((x) % 4 == 0 ? x : (((x) + 3) & (~0x03)))

/*
 * Se ocupa de reservar/desreservar bloques, actualizando las estructuras del
 * filesystem que indican cuantos y cuales bloques estan en uso.
 * reserve = true, reserva bloques
 * reserve = false, desreserva bloques
 *
 */
uint32_t ext2_set_resvd_blocks(fs_info* info, uint32_t from, uint32_t count,
                               bool reserve) {
    block_cache* cache = info->cache;
    uint32_t bn_per_group = info->sb->blocks_per_group;
    uint64_t block_shift = info->block_shift;
    uint64_t first_block = (block_shift == EXT2_MIN_BLOCK_SHIFT);
    uint64_t block_group = (from - first_block) / bn_per_group;
    uint64_t block_offset = (from - first_block) % bn_per_group;
    uint64_t b_size_bits = info->block_size * 8;
    // block_bitmap apunta al bloque con el bitmap del block usage del grupo
    void* block_bitmap =
        block_cache_load_dirty(cache, info->gt[block_group].bn_block_bitmap);
    // Seteo los bits del bitmap que corresponden
    if (reserve) {
        bitmap_set_bits(block_bitmap, b_size_bits, block_offset, count);
        info->sb->free_blocks -= count;
        info->gt[block_group].free_blocks -= count;
    } else {
        bitmap_clear_bits(block_bitmap, b_size_bits, block_offset, count);
        info->sb->free_blocks += count;
        info->gt[block_group].free_blocks += count;
    }
    block_cache_unload(cache, block_bitmap);

    return 0;
}

/* ext2_new_group
 * Esta funcion se encarga de inaugurar nuevos grupos de bloques. Debe tener
 * en cuenta no pasarse del tamano del drive y mantener el superbloque esparso
 * si asi corresponde.
 */
uint64_t ext2_new_group(fs_info* info) {
    return 0;
    /* Aca hay varios temas de especificacion de ext2 que estuvimos discutiendo
     * con noit que son muy raros. No se aclaran ciertos casos bordes al crear
     * nuevos grupos de bloques, entonces prefiero no hacerlo por ahora.
     * Tambien hay que tener en cuenta no pasarnos del tamano del drive al crear
     * un nuevo grupo, cosa que es imposible saber por ahora (el driver de
     * disco por ahora no puede obtener esta informacion), entonces prefiero
     * dejarlo para despues. TODO.
     * De todos modos, esta funcion se usa en casos muy borde, cuando nos
     * quedamos sin el espacio actual alocado en todo el filesystem. De hecho
     * la especificacion ni siquiera es clara si esta operacion se puede hacer
     * mientras el disco esta online y montado.
     */
    /*
    uint64_t block_shift = info->block_shift;
    uint64_t block_size = 1 << block_shift;
    uint32_t bn_per_group = info->sb->blocks_per_group;

    block_cache* cache = info->cache;

    // http://www.nongnu.org/ext2-doc/ext2.html seccion 3.1.6
    uint64_t first_block = (block_shift == EXT2_MIN_BLOCK_SHIFT);

    uint64_t group = info->group_count++;
    uint32_t first_block_in_group = first_block + group * bn_per_group;
    info->sb->group_count++;
    info->sb->number_blocks += bn_per_group;

    ext2_group_desc* group_desc = &info->gt[group];

    uint32_t offset = 0;

    if(ext2_has_super(info, group)){
        void* superblock = block_cache_load(cache, first_block_in_group +
    offset);
        memmove(superblock, info->sb, block_size);
        block_cache_unload(cache, superblock);
        offset++;
        void* group_dt = block_cache_load(cache, first_block_in_group + offset);
        memmove(group_dt, info->gt, block_size);
        block_cache_unload(cache, superblock);
        offset++;
    }
    */
}

/* ext2_balloc
 *
 * Intenta alocar una constante fija de bloques para el archivo dado.
 * La constante esta dada por PREALLOC_BLOCK_COUNT. Se podria usar tambien
 * la constante que viene dada en el superbloque, pero es opcional.
 * El algoritmo es simple  e intenta minimizar la fragmentacion externa:
 * intenta alocar desde el ultimo bloque utilizado por el archivo en adelante.
 * Si no encuentra en el grupo actual, sigue con los grupos siguientes.
 */
uint32_t ext2_balloc(fs_info* info, fs_file* file) {
    // No se alocaron bloques todavia, tengo que hacerlo, para eso cargo
    // el group descriptor
    uint64_t block_shift = info->block_shift;
    uint32_t bn_per_group = info->sb->blocks_per_group;
    uint32_t block_number = file->block_number;
    block_cache* cache = info->cache;
    // Si el archivo esta vacio, block_number vale 0, entonces lo seteo en 1
    // para que funcione todo correctamente.
    if (block_number == 0)
        block_number = 1;
    // http://www.nongnu.org/ext2-doc/ext2.html seccion 3.1.6
    uint64_t first_block = (block_shift == EXT2_MIN_BLOCK_SHIFT);
    uint64_t block_group_num = (block_number - first_block) / bn_per_group;
    // Si ya tenia bloques prealocados, lo limpio
    if (file->prealloc_count > 0) {
        ext2_set_resvd_blocks(info, file->prealloc_block, file->prealloc_count,
                              false);
        file->prealloc_count = 0;
        file->prealloc_block = 0;
    }
    for (uint64_t group = block_group_num; group <= info->group_count;
         group++) {
        uint32_t block_offset_in_group;
        if (group == block_group_num)
            block_offset_in_group = (block_number - first_block) % bn_per_group;
        else
            block_offset_in_group = 0;
        uint32_t first_block_in_group = first_block + group * bn_per_group;
        // block_bitmap apunta al bloque con el bitmap del block usage del grupo
        void* block_bitmap =
            block_cache_load(cache, info->gt[group].bn_block_bitmap);
        // Obtengo el bloque buscando 0's consecutivos y sumandole el offset
        // correspondiente
        uint64_t block =
            first_block_in_group +
            bitmap_search_zeros(block_bitmap, info->block_size * 8,
                                block_offset_in_group, PREALLOC_BLOCK_COUNT);
        block_cache_unload(cache, block_bitmap);
        // Si el bloque que encontre esta en rango, listo
        if (block - first_block_in_group < info->block_size * 8) {
            ext2_set_resvd_blocks(info, block, PREALLOC_BLOCK_COUNT, true);
            file->prealloc_block = block;
            file->prealloc_count = PREALLOC_BLOCK_COUNT;
            break;
        }
        // Ultima chance para obtener un bloque nuevo, crear un grupo nuevo
        if (group == info->group_count - 1) {
            uint64_t new_group = ext2_new_group(info);
            if (new_group == 0)
                return 0;
        }
    }
    return 0;
}

/* ext2_get_next_prealloc
 * Devuelve el proximo bloque de los prealocados por el archivo
 */
uint32_t ext2_get_next_prealloc(fs_file* file) {
    uint32_t block = file->prealloc_block;
    file->prealloc_block++;
    file->prealloc_count--;
    return block;
}

/* Cuenta cuantos bloques reales necesito para devolver un bloque real de
 * datos.
 */
uint64_t ext2_needed_blocks(fs_info* info, fs_file* file) {
    uint64_t block_shift = info->block_shift;
    uint64_t bn_per_block = 1 << (block_shift - 2);
    uint64_t index = file->block_index;
    uint64_t how_many = 1;
    // Estoy en los primeros bloques, los directos de 0 a 11.
    if (index < 12)
        return how_many;
    // Me fijo si necesito un simple indirecto
    if ((index - 12) % bn_per_block == 0)
        how_many++;
    // Me fijo si necesito un doble indirecto
    if (index - 12 >= bn_per_block &&
        (index - 12 - bn_per_block) % (bn_per_block * bn_per_block) == 0)
        how_many++;
    // Me fijo si necesito el triple indirecto
    if (index == 12 + bn_per_block + bn_per_block * bn_per_block)
        how_many++;
    return how_many;
}

/* ext2_assign_blocks
 *
 * Arma la estructura de arbol del inodo como corresponda para expandir el
 * tamano efectivo del archivo en 1 bloque. Para esto, tiene que eventualmente
 * escribir nuevos bloques indirectos simples, dobles o triples.
 */
uint32_t ext2_assign_blocks(fs_info* info, fs_file* file) {
    block_cache* cache = info->cache;
    uint64_t bn_per_block = 1 << (info->block_shift - 2);
    uint64_t index = file->block_index;
    // Asigno el bloque que necesito, y despues procedo a rearmar
    // La estructura de ext2 para que quede todo coherenet
    file->block_number = ext2_get_next_prealloc(file);
    // Estoy en los primeros bloques, los directos de 0 a 11.
    if (index < 12) {
        file->inode->bn_direct[file->block_index] = file->block_number;
        return 0;
    }
    // Calculo los nuevos indices
    uint64_t new_single = ((index - 12) >> (info->block_shift - 2)) + 1;
    uint64_t new_double = ((new_single - 2) >> (info->block_shift - 2)) + 1;
    // Me fijo si necesito un nuevo single
    if ((index - 12) % bn_per_block == 0) {
        uint32_t new_single_block = ext2_get_next_prealloc(file);
        // Actualizo la estructura del inodo, si es necesario
        if (index == 12)
            file->inode->bn_single = new_single_block;
        // Si tenia un single viejo, lo descargo y cargo el nuevo
        if (file->singly_indirect != NULL)
            block_cache_unload(cache, file->singly_indirect);
        file->singly_indirect = block_cache_load_dirty(cache, new_single_block);
        // Me fijo si necesito un nuevo doble
        if ((index - 12 >= bn_per_block &&
             (new_single - 2) % bn_per_block == 0)) {
            uint32_t new_double_block = ext2_get_next_prealloc(file);
            // Actualizo la estructura del inodo, si es necesario
            if (index == bn_per_block + 12)
                file->inode->bn_double = new_double_block;
            if (file->doubly_indirect != NULL)
                block_cache_unload(cache, file->doubly_indirect);
            file->doubly_indirect =
                block_cache_load_dirty(cache, new_double_block);
            // Me fijo si necesito un nuevo triple
            if (index == 12 + bn_per_block + bn_per_block * bn_per_block) {
                uint32_t new_triple_block = ext2_get_next_prealloc(file);
                file->inode->bn_triple = new_triple_block;
                // Seria muy raro que pase esto, "habia otro triple cargado"
                if (file->triply_indirect != NULL)
                    block_cache_unload(cache, file->triply_indirect);
                file->triply_indirect =
                    block_cache_load_dirty(cache, new_triple_block);
            }
            if (new_double > 1)
                file->triply_indirect[(new_double - 2) % bn_per_block] =
                    new_double_block;
        }
        if (new_single > 1)
            file->doubly_indirect[(new_single - 2) % bn_per_block] =
                new_single_block;
    }
    file->singly_indirect[(index - 12) % bn_per_block] = file->block_number;
    block_cache_set_dirty(cache, file->singly_indirect);
    return 0;
}

/*
 * Le desasigna el bloque actual al archivo, que deberia ser el ultimo
 */
uint32_t ext2_unassign_block(fs_info* info, fs_file* file) {
    block_cache* cache = info->cache;
    uint64_t bn_per_block = 1 << (info->block_shift - 2);
    uint64_t index = file->block_index;
    // Estoy en los primeros bloques, los directos de 0 a 11.
    if (index < 12) {
        if (file->inode->bn_direct[index] == 0)
            return 0;
        ext2_set_resvd_blocks(info, file->inode->bn_direct[index], 1, false);
        file->inode->bn_direct[index] = 0;
        // Sectores de 512 bytes usados
        file->inode->used_sectors -= info->block_size / 512;
        return 0;
    }
    uint64_t single_index = ((index - 12) >> (info->block_shift - 2)) + 1;
    uint64_t double_index = ((single_index - 2) >> (info->block_shift - 2)) + 1;
    uint64_t how_many = 1;
    ext2_set_resvd_blocks(info, file->block_number, 1, false);
    // Me fijo si tengo que borrar un single
    if ((index - 12) % bn_per_block == 0) {
        uint32_t single_block;
        if (index == 12) {
            single_block = file->inode->bn_single;
            file->inode->bn_single = 0;
        } else {
            single_block = file->doubly_indirect[single_index - 2];
        }
        ext2_set_resvd_blocks(info, single_block, 1, false);
        how_many++;
        // Si tenia un single viejo, lo descargo
        if (file->singly_indirect != NULL) {
            block_cache_unload(cache, file->singly_indirect);
            file->singly_indirect = NULL;
        }
        // Me fijo si hay que borrar el doble
        if ((index - 12 >= bn_per_block &&
             (single_index - 2) % bn_per_block == 0)) {
            // Actualizo la estructura del inodo, si es necesario
            uint32_t double_block;
            if (index == bn_per_block + 12) {
                double_block = file->inode->bn_double;
                file->inode->bn_double = 0;
            } else {
                double_block = file->triply_indirect[double_index - 2];
            }
            ext2_set_resvd_blocks(info, double_block, 1, false);
            how_many++;
            if (file->doubly_indirect != NULL) {
                block_cache_unload(cache, file->doubly_indirect);
                file->doubly_indirect = NULL;
            }
            // Me fijo si hay que borrar el triple
            if (index == 12 + bn_per_block + bn_per_block * bn_per_block) {
                ext2_set_resvd_blocks(info, file->inode->bn_triple, 1, false);
                file->inode->bn_triple = 0;
                how_many++;
                if (file->triply_indirect != NULL) {
                    block_cache_unload(cache, file->triply_indirect);
                    file->triply_indirect = NULL;
                }
            }
        }
    }
    // Sectores de 512 bytes usados
    file->inode->used_sectors -= how_many * info->block_size / 512;
    return 0;
}

/* ext2_request_block
 *
 * se ocupa de calcular cuantos bloques son necesarios para brindar un
 * bloque de datos, actualiza el inodo y asigna los bloques, armando
 * correctamente la estructura de arbol del inodo de ext2
 */
uint32_t ext2_request_block(fs_info* info, fs_file* file) {
    uint64_t bn_per_block = 1 << (info->block_shift - 2);
    uint64_t max_file_size = 12 + bn_per_block + bn_per_block * bn_per_block +
                             bn_per_block * bn_per_block * bn_per_block;
    if (file->block_index == max_file_size)
        return 1;
    // Calculamos cuantos bloques necesitamos para tener un bloque real de data
    uint64_t how_many = ext2_needed_blocks(info, file);
    if (file->prealloc_block == 0 || file->prealloc_count < how_many) {
        ext2_balloc(info, file);
        // Nos fijamos si ahora si le alcanza.
        if (file->prealloc_block == 0 || file->prealloc_count < how_many)
            KERNEL_PANIC("EXT2: Cannot request block\n");
    }
    // Sectores de 512 bytes usados
    file->inode->used_sectors += how_many * info->block_size / 512;
    // Asignamos los bloques, si hay error, devuelve el error
    return ext2_assign_blocks(info, file);
}

uint64_t fs_write(fs_info* info, fs_file* file, const void* buffer,
                  uint64_t count) {
    block_cache_set_dirty(info->head_cache, info->sb);
    block_cache_set_dirty(info->head_cache, info->gt);
    block_cache* cache = info->cache;
    // remaining_after_eof es el numero de bytes que hay despues de que el
    // archivo  termina, porque no esta usando todo el bloque.
    // Cuando file->block_offset == block_size-1,
    // quiero que remaining_after_eof valga 0
    uint64_t remaining_after_eof =
        (-file->remaining - file->block_offset) % info->block_size;
    // Escribo primero hasta el final del ultimo bloque, porque seguro
    // puedo hacerlo. fst_count es eso, la cantidad de bytes que puedo
    // escribir hasta quedarme sin bloques.
    uint64_t fst_count = MIN(count, file->remaining + remaining_after_eof);
    uint64_t written = 0;
    uint64_t size = count;
    while (size) {
        // block_size - file->block_offset es lo que me queda de un bloque.
        uint64_t to_copy = MIN(size, info->block_size - file->block_offset);
        // si me pase de lo que tenia alocado, tengo que pedir mas
        if (written >= fst_count) {
            // pido un bloque, si devuelve distinto de 0, salgo
            if (ext2_request_block(info, file))
                break;
        }
        void* block_ptr = block_cache_load_dirty(cache, file->block_number);
        void* base = (void*)((uintptr_t)block_ptr + file->block_offset);
        memcpy(base, buffer, to_copy);
        block_cache_unload(cache, block_ptr);
        // Me fijo cuanto tengo que correr el eof (file->remaining)
        if (file->remaining < to_copy) {
            uint32_t eof_advanced = to_copy - file->remaining;
            block_cache_set_dirty(cache, file->inode);
            // Actualizo el tamano del inodo, me fijo primero si hay overflow
            if (file->inode->size_low + eof_advanced < file->inode->size_low)
                file->inode->size_high++;
            file->inode->size_low += eof_advanced;
            file->remaining += eof_advanced;
        }
        ext2_seek(cache, info->block_shift, file, to_copy);
        buffer = (void*)((uintptr_t)buffer + to_copy);
        size -= to_copy;
        written += to_copy;
    }
    return written;
}

/*
 * La funcion que sigue es analoga a la de bloques, se ocupa de actualizar las
 * estructuras del filesystem que indican cuantos y cuales inodos estan en uso
 */
uint32_t ext2_set_resvd_inodes(fs_info* info, uint32_t from, uint32_t count,
                               bool reserve) {
    uint32_t in_per_group = info->sb->inodes_per_group;
    uint32_t group = (from - 1) / in_per_group;
    uint32_t inode_offset = (from - 1) % in_per_group;
    // inode_bitmap apunta al bloque con el bitmap de inode usage del grupo
    void* inode_bitmap =
        block_cache_load_dirty(info->cache, info->gt[group].bn_inode_bitmap);
    if (reserve) {
        bitmap_set_bits(inode_bitmap, in_per_group, inode_offset, count);
        info->sb->free_inodes -= count;
        info->gt[group].free_inodes -= count;
    } else {
        bitmap_clear_bits(inode_bitmap, in_per_group, inode_offset, count);
        info->sb->free_inodes += count;
        info->gt[group].free_inodes += count;
    }
    block_cache_unload(info->cache, inode_bitmap);
    return 0;
}

uint32_t ext2_ialloc(fs_info* info, fs_dir* parent) {
    uint32_t in_per_group = info->sb->inodes_per_group;
    // Podria elegir empezar a partir de cualquier grupo, tome la decision
    // de empezar desde 0. Otra opcion: (inode_number - 1) / in_per_group;
    uint32_t group_number = 0;
    block_cache* cache = info->cache;
    for (uint64_t group = group_number; group <= info->group_count; group++) {
        // inode_bitmap apunta al bloque con el bitmap de inode usage del grupo
        void* inode_bitmap =
            block_cache_load(cache, info->gt[group].bn_inode_bitmap);
        // Obtengo el bloque, buscando 0's consecutivos y sumandole el offset
        // correspondiente
        uint32_t first_inode_in_group = 1 + group * in_per_group;
        uint64_t inode = first_inode_in_group +
                         bitmap_search_zeros(inode_bitmap, in_per_group, 0, 1);
        block_cache_unload(cache, inode_bitmap);
        if (inode - first_inode_in_group < in_per_group) {
            ext2_set_resvd_inodes(info, inode, 1, true);
            return inode;
        }
        // Ultima chance, crear un grupo nuevo
        if (group == info->group_count - 1) {
            uint64_t new_group = ext2_new_group(info);
            if (new_group == 0)
                return 0;
        }
    }
    return 0;
}

int32_t ext2_add_direntry(fs_info* info, fs_dir* dir, char* filename,
                          uint32_t inode, uint16_t filetype) {
    uint16_t name_length = strlen(filename);
    // Tiene que ser multiplos de 4
    uint16_t reclen = MULTIPLE_OF_4(sizeof(ext2_dir_entry) + name_length);
    // Armo la entrada de la carpeta
    uint8_t direntry_filetype;
    // http://www.nongnu.org/ext2-doc/ext2.html#IFDIR-FILE-TYPE
    switch (filetype) {
        case FILE_TYPE_REGULAR_FILE:
            direntry_filetype = 1;
            break;
        case FILE_TYPE_DIRECTORY:
            direntry_filetype = 2;
            break;
        default:
            direntry_filetype = 0;
            break;
    }
    ext2_dir_entry entry = {inode, reclen, name_length, direntry_filetype};
    // Lo seteo dirty porque lo voy a modificar
    block_cache_set_dirty(info->cache, dir->file.inode);
    ext2_dir_entry curr_entry;
    curr_entry.inode_number = 0;
    // Buscamos la primera entry no nula
    while (fs_read(info, &dir->file, &curr_entry, sizeof(ext2_dir_entry))) {
        uint32_t next_entry = curr_entry.rec_len - sizeof(ext2_dir_entry);
        if (curr_entry.inode_number) {
            // si encontre, me atraso para poder volver a leerlo en el while
            fs_seek(info, &dir->file, -sizeof(ext2_dir_entry));
            break;
        }
        fs_seek(info, &dir->file, next_entry);
    }
    // Calculo el tamano del archivo
    uint64_t size = ext2_get_size(dir->file.inode);
    // Ciclo por todo el archivo del directorio buscando un lugar libre
    while (1) {
        fs_read(info, &dir->file, &curr_entry, sizeof(ext2_dir_entry));
        uint64_t curr_entry_size = sizeof(ext2_dir_entry) + curr_entry.name_len;
        // Me fijo si entra despues del actual
        if (curr_entry.rec_len - curr_entry_size > entry.rec_len) {
            uint16_t new_curr_rec_len = MULTIPLE_OF_4(curr_entry_size);
            uint16_t diff = new_curr_rec_len - curr_entry_size;
            entry.rec_len = curr_entry.rec_len - new_curr_rec_len;
            curr_entry.rec_len = new_curr_rec_len;
            // Vuelvo para sobreescribir la anterior entrada
            fs_seek(info, &dir->file, -sizeof(ext2_dir_entry));
            // Sobreescribo la anterior entrada (el nombre lo salteo)
            fs_write(info, &dir->file, &curr_entry, sizeof(ext2_dir_entry));
            fs_seek(info, &dir->file, curr_entry.name_len + diff);
            // Escribo la nueva entrada
            fs_write(info, &dir->file, &entry, sizeof(ext2_dir_entry));
            fs_write(info, &dir->file, filename, entry.name_len);
            break;
        } else if (dir->file.remaining - curr_entry.name_len < entry.rec_len) {
            // Estoy en la ultima entrada del bloque
            // La distancia al proximo es block_size
            entry.rec_len = info->block_size;
            // Voy hasta el final del archivo
            fs_seek(info, &dir->file, dir->file.remaining);
            // Escribo la nueva entrada
            fs_write(info, &dir->file, &entry, sizeof(ext2_dir_entry));
            fs_write(info, &dir->file, filename, entry.name_len);
            // Actualizo el tamano
            size += info->block_size;
            break;
        } else {
            fs_seek(info, &dir->file,
                    curr_entry.rec_len - sizeof(ext2_dir_entry));
        }
    }
    dir->file.inode->size_low = size;
    dir->file.inode->size_high = size >> 32;
    return 0;
}

int64_t fs_create(fs_info* info, fs_dir* dir, dir_entry* file_info) {
    uint32_t inode = ext2_ialloc(info, dir);
    // Creo la entrada
    ext2_add_direntry(info, dir, file_info->name, inode, file_info->file_type);
    // Le cargo toda la informacion al inodo
    ext2_inode* inode_ptr = ext2_load_inode(info, inode);
    block_cache_set_dirty(info->cache, (void*)inode_ptr);
    // Pongo todo en 0
    memset(inode_ptr, 0, sizeof(ext2_inode));
    inode_ptr->hard_links = 1;
    inode_ptr->mode = 0000644;
    switch (file_info->file_type) {
        case FILE_TYPE_REGULAR_FILE:
            inode_ptr->mode |= 0x8000;
            inode_ptr->size_low = 0;
            inode_ptr->size_high = 0;
            break;
        case FILE_TYPE_DIRECTORY:
            // Actualizo hard link count
            dir->file.inode->hard_links++;
            inode_ptr->hard_links++;
            // Actualizo cantidad de inodos destinados a directorio del grupo
            block_cache_set_dirty(info->head_cache, info->gt);
            uint32_t in_per_group = info->sb->inodes_per_group;
            uint32_t group = (dir->file.inode_number - 1) / in_per_group;
            info->gt[group].dir_inodes++;
            // Creo propiamente el directorio
            inode_ptr->mode |= 0x4000;
            uint32_t dir_size = info->block_size;
            uint8_t zeros[dir_size];
            for (uint64_t i = 0; i < dir_size; i++)
                zeros[i] = 0;
            fs_dir new_dir;
            ext2_load_dir(inode_ptr, &new_dir, inode);
            fs_write(info, &new_dir.file, zeros, dir_size);
            fs_seek(info, &new_dir.file, -(int64_t)dir_size);
            ext2_dir_entry entry = {inode, dir_size, 1, 2};
            fs_write(info, &new_dir.file, &entry, sizeof(ext2_dir_entry));
            fs_write(info, &new_dir.file, ".", 1);
            fs_seek(info, &new_dir.file, -sizeof(ext2_dir_entry) - 1);
            ext2_add_direntry(info, &new_dir, "..", dir->file.inode_number,
                              FILE_TYPE_DIRECTORY);
            ext2_unload_file(info, &new_dir.file);
            break;
    }
    block_cache_unload(info->cache, (void*)inode_ptr);
    return 0;
}

int64_t fs_truncate(fs_info* info, fs_file* file, uint64_t length) {
    block_cache_set_dirty(info->cache, file->inode);
    // voy al final del archivo
    fs_seek(info, file, file->remaining - 1);
    uint64_t size = ext2_get_size(file->inode), size_copy = size;
    while ((length >> info->block_shift) < (size >> info->block_shift)) {
        ext2_unassign_block(info, file);
        size -= info->block_size;
        fs_seek(info, file, -info->block_size);
    }
    size = length;
    // caso borde, si tengo que borrar todo el bloque
    if ((size & (info->block_size - 1)) == 0)
        ext2_unassign_block(info, file);
    file->inode->size_low = size;
    file->inode->size_high = size >> 32;
    ext2_seek(info->cache, info->block_shift, file, -size_copy + 1);
    file->remaining = length;
    return 0;
}

int64_t fs_remove(fs_info* info, fs_dir* dir) {
    if (!dir || !dir->rec_inode)
        return -1;
    // Retrocedo tanto rec_len, porque quiero sobreescribir la entrada
    ext2_seek(info->cache, info->block_shift, &dir->file, -dir->rec_len);
    ext2_set_resvd_inodes(info, dir->rec_inode, 1, false);
    // Seteo el count de hard_links del archivo en 0.
    // TODO: setear dtime segun corresponda
    fs_file deleted_file;
    fs_open(info, dir, &deleted_file);
    fs_truncate(info, &deleted_file, 0);
    block_cache_set_dirty(info->cache, deleted_file.inode);
    deleted_file.inode->hard_links = 0;
    deleted_file.inode->deletion_time = 1;
    switch (deleted_file.inode->mode & 0xf000) {
        case 0x8000:  // regular file
            break;
        case 0x4000:  // directory
            // actualizo hard link count
            dir->file.inode->hard_links--;
            // actualizo cantidad de inodos destinados a directorio del grupo
            block_cache_set_dirty(info->head_cache, info->gt);
            uint32_t in_per_group = info->sb->inodes_per_group;
            uint32_t group = (dir->file.inode_number - 1) / in_per_group;
            info->gt[group].dir_inodes--;
            break;
    }
    fs_close(info, &deleted_file);
    // Escribo un 0 en donde deberia estar el numero de inodo del archivo
    block_cache_set_dirty(info->cache, dir->file.inode);
    uint32_t inode_0 = 0;
    fs_write(info, &dir->file, (void*)&inode_0, sizeof(inode_0));
    dir->rec_inode = 0;
    // TODO: falta el caso borde en que borro el ultimo archivo del directorio
    // y ademas era el unico del bloque, entonces tengo que liberar el bloque
    // Idea: llamar a truncate y que haga todo el trabajo por mi
    // Esto es opcional
    return 0;
}

/* ext2_has_super
 *
 * @brief Dice si un grupo del filesystem debe contener o no un backup del
 * superbloque y el group descriptor table.
 *
 * @param info la informacion del filesystem
 * @param group el numero de grupo del grupo en cuestion
 *
 * @returns 1 si tiene que tener backup, 0 si no
 */
uint32_t ext2_has_super(const fs_info* info, uint64_t group) {
    // Tenemos que chequear si group es potencia de 3, 5 o 7.
    // Asi lo hace el kernel de linux, no voy a hacer algo mas sofisticado
    if (group == 0)
        return 1;
    if (info->sb->req_write_flags & EXT2_FEATURE_SPARSE)
        return 0;
    uint64_t a = 1, b = 1, c = 1;
    while (a < group)
        a *= 3;
    while (b < group)
        b *= 5;
    while (c < group)
        c *= 7;
    return a == group || b == group || c == group;
}

/* ext2_restore_backup
 *
 * @brief Escribe el superbloque y la group descriptor table en todos los grupos
 * que deben tener el backup.
 *
 * @param info la informacion del filesystem
 *
 * @returns 0 si no hubo error
 */
uint32_t ext2_restore_backup(const fs_info* info) {
    block_cache* cache = info->cache;
    uint64_t first_block = (info->block_shift == EXT2_MIN_BLOCK_SHIFT);
    uint32_t bn_per_group = info->sb->blocks_per_group;
    // Arranco desde 1 porque el superbloque del grupo 0 se actualiza siempre
    for (uint64_t group = 1; group < info->group_count; group++) {
        uint64_t first_block_grp = first_block + group * bn_per_group;
        if (ext2_has_super(info, group)) {
            // cargo el superbloque del grupo
            void* superblock = block_cache_load_dirty(cache, first_block_grp);
            memcpy(superblock, info->sb, info->block_size);
            block_cache_unload(cache, superblock);
            // cargo la tabla de descriptores de grupos
            void* groupdt = block_cache_load_dirty(cache, first_block_grp + 1);
            memcpy(groupdt, info->gt, info->block_size);
            block_cache_unload(cache, groupdt);
        }
    }
    return 0;
}
