#include <exception.h>
#include <hdd.h>
#include <i386.h>
#include <macros.h>
#include <ports.h>
#include <print.h>
#include <timer.h>

// El filesystem se construye a partir del primer MiB
#define FS_SECTOR 2048

static const char* ide_type_names[] = {"ATA", "ATAPI", "SATA", "SATAPI"};

static void hdd_ata_wait(uint16_t port) {
    while (FLAGS_SET(inb(ATA_COMSTAT + port), ATA_STATUS_BSY));
}

static uint8_t ata_read_stable(uint16_t port) {
    uint8_t status;
    // Se lee el registro 5 veces para generar un
    // delay de 400 ns de acuerdo a la especificacion
    status = inb(ATA_COMSTAT + port);
    status = inb(ATA_COMSTAT + port);
    status = inb(ATA_COMSTAT + port);
    status = inb(ATA_COMSTAT + port);
    status = inb(ATA_COMSTAT + port);
    return status;
}

static void hdd_setup_lba(uint32_t lba_address, uint32_t sectors, char cmd) {
    inb(ATA_PRIMARY_ERR);
    hdd_ata_wait(ATA_PRIMARY_BASE);
    while (!FLAGS_SET(inb(ATA_PRIMARY_COMSTAT), ATA_STATUS_RDY))
        ;
    // Ya puedo pedirle al disco
    outb(ATA_PRIMARY_DEVCONTROL, 0x0A);
    outb(ATA_PRIMARY_ERR, 0x00);
    outb(ATA_PRIMARY_SECTORCOUNT, sectors);
    outb(ATA_PRIMARY_LBALOW, (char)lba_address);
    outb(ATA_PRIMARY_LBAMID, (char)(lba_address >> 8));
    outb(ATA_PRIMARY_LBAHIGH, (char)(lba_address >> 16));
    outb(ATA_PRIMARY_DRIVE, 0xE0 | ((lba_address >> 24) & 0x0F));
    outb(ATA_PRIMARY_COMSTAT, cmd);
}

void hdd_read(uint32_t lba_address, uint32_t sectors, void* buffer) {
    outb(ATA_DRIVE + ATA_PRIMARY_BASE, 0xA0);
    hdd_ata_wait(ATA_PRIMARY_BASE);
    uint16_t* word_buffer = buffer;
    hdd_setup_lba(lba_address + FS_SECTOR, sectors, ATA_COMM_READ_LBA28);
    uint64_t flags = cli_safe();
    for (uint32_t i = 0; i < sectors; i++) {
        hdd_ata_wait(ATA_PRIMARY_BASE);
        for (uint32_t j = 0; j < ATA_SECTOR_WORDS; j++)
            *(word_buffer++) = inw(ATA_PRIMARY_DATA);
        ata_read_stable(ATA_PRIMARY_BASE);
    }
    sti_safe(flags);
}

void hdd_write(uint32_t lba_address, uint32_t sectors, const void* buffer) {
    outb(ATA_DRIVE + ATA_PRIMARY_BASE, 0xA0);
    hdd_ata_wait(ATA_PRIMARY_BASE);
    const uint16_t* word_buffer = buffer;
    hdd_setup_lba(lba_address + FS_SECTOR, sectors, ATA_COMM_WRITE_LBA28);
    uint64_t flags = cli_safe();
    for (uint32_t i = 0; i < sectors; i++) {
        hdd_ata_wait(ATA_PRIMARY_BASE);
        for (uint32_t j = 0; j < ATA_SECTOR_WORDS; j++)
            outw(ATA_PRIMARY_DATA, *(word_buffer++));
        hdd_ata_wait(ATA_PRIMARY_BASE);
        outb(ATA_PRIMARY_COMSTAT, ATA_COMM_FLUSHCACHE);
    }
    hdd_ata_wait(ATA_PRIMARY_BASE);
    sti_safe(flags);
}

struct ide_channels {
    uint16_t base;      // Puerto base
    uint16_t ctrl;      // Puerto de control Base
    uint16_t bmide;     // Bus Master IDE
    uint8_t nIEN;       // No interrupt
} channels[2];

struct ide_device {
    uint8_t resvd;      // 0:vacio, 1:existe
    uint8_t channel;    // 0:primary, 1:secondary
    uint8_t drive;      // 0:master, 1:slave
    uint16_t type;
    uint16_t signature;
    uint16_t features;
    uint64_t command_sets;
    uint64_t size;      // Tamaño en bytes
    char model[41];     // Modelo (string)
} ide_devices[4];

uint8_t hdd_init(uint16_t port, uint8_t slave, uint8_t dev) {
    // http://wiki.osdev.org/ATA_PIO_Mode#IDENTIFY_command
    uint8_t type = IDE_ATA;
    identify_data identify;
    // El dispositivo puede no existir, asi que no podemos leer directamente.
    // Primero, lo seleccionamos con el outb, esperamos 1/4 de segundo (mas que
    // suficiente) y despues leemos el campo de status del dispositivo.
    // Si vale 0xFF, en particular esta BSY y DRQ, cosa que no puede pasar,
    // entonces el dispositivo no existe.
    outb(ATA_DRIVE + port, slave ? 0xB0 : 0xA0);
    sleep(256);
    if (inb(ATA_COMSTAT + port) == 0xFF)
        return 0;
    hdd_ata_wait(port);
    outb(ATA_SECTORCOUNT + port, 0);
    outb(ATA_LBALOW + port, 0);
    outb(ATA_LBAMID + port, 0);
    outb(ATA_LBAHIGH + port, 0);
    // mando un request de IDENTIFY
    outb(ATA_COMSTAT + port, 0xEC);
    uint8_t result = ata_read_stable(port), is_ata = 1;
    // Si es cero no se ha detectado un disco
    if (!result)
        return result;
    hdd_ata_wait(port);
    // Segun osdev, hasta aca es el standard de ata, pero hay devices que
    // no siguen la spec, entonces hay que hacer lo que sigue
    if (inb(ATA_LBAMID + port) && inb(ATA_LBAHIGH + port))
        is_ata = 0;  // Se detecto un dispositivo master, pero no es ATA
    // Tengo que pollear hasta que se setee DRQ o ERR
    uint8_t status;
    do {
        status = inb(ATA_COMSTAT + port);
    } while ((status & (ATA_STATUS_ERR | ATA_STATUS_DRQ)) == 0);
    // Leemos la informacion que nos brinda IDENTIFY
    for (uint64_t i = 0; i < ATA_SECTOR_WORDS; i++)
        ((uint16_t*)&identify)[i] = inw(ATA_DATA + port);
    // Chequeo si es ATAPI
    if (!is_ata) {
        uint8_t cl = inb(port + ATA_LBAMID);
        uint8_t ch = inb(port + ATA_LBAHIGH);
        // http://wiki.osdev.org/ATA_PIO_Mode#Detecting_device_types
        if (cl == 0x14 && ch == 0xEB)
            type = IDE_ATAPI;
        else if (cl == 0x69 && ch == 0x96)
            type = IDE_SATAPI;
        else if (cl == 0x3C && ch == 0xC3)
            type = IDE_SATA;
        else
            return 0;  // Unknown type (may not be a device).
    }
    // Cargamos la informacion de identify
    ide_devices[dev].type = type;
    ide_devices[dev].signature = identify.device_type;
    ide_devices[dev].features = identify.capabilities;
    ide_devices[dev].command_sets = identify.command_sets;
    // Obtengo el tamano del disco
    if (ide_devices[dev].command_sets & (1 << 26))
        // 48-Bit LBA
        ide_devices[dev].size = identify.max_lba_ext;
    else
        // CHS o 28-bit LBA
        ide_devices[dev].size = identify.max_lba;
    ide_devices[dev].size *= ATA_SECTOR_SIZE;
    // Modelo del disco (40 chars)
    for (uint64_t k = 0; k < 40; k++)
        ide_devices[dev].model[k] = identify.model[k];
    ide_devices[dev].model[40] = '\0';
    return 1;
}

void ide_init(uint32_t bar0, uint32_t bar1, uint32_t bar2, uint32_t bar3,
              uint32_t bar4) {
    // Detecto los puertos de I/O que interfacean al controlador IDE
    // http://wiki.osdev.org/PCI_IDE_Controller#Detecting_IDE_Drives
    channels[ATA_PRIMARY  ].base  = (bar0 & 0xFFFFFFFC) + 0x1F0 * (!bar0);
    channels[ATA_PRIMARY  ].ctrl  = (bar1 & 0xFFFFFFFC) + 0x3F6 * (!bar1);
    channels[ATA_SECONDARY].base  = (bar2 & 0xFFFFFFFC) + 0x170 * (!bar2);
    channels[ATA_SECONDARY].ctrl  = (bar3 & 0xFFFFFFFC) + 0x376 * (!bar3);
    channels[ATA_PRIMARY  ].bmide = (bar4 & 0xFFFFFFFC) + 0;  // Bus Master IDE
    channels[ATA_SECONDARY].bmide = (bar4 & 0xFFFFFFFC) + 8;  // Bus Master IDE
    // Apago las interrupciones
    outb(ATA_PRIMARY_BASE   + ATA_CONTROL, ATA_CTRL_NIEN);
    outb(ATA_SECONDARY_BASE + ATA_CONTROL, ATA_CTRL_NIEN);
    // Detecto los dispositivos
    for (int i = 0, count = 0; i < 2; i++) {
        for (int j = 0; j < 2; j++, count++) {
            ide_devices[count].resvd = hdd_init(channels[i].base, j, count);
            ide_devices[count].channel = i;
            ide_devices[count].drive = j;
        }
    }
    // Imprimo el resultado
    for (int i = 0; i < 4; i++)
        if (ide_devices[i].resvd == 1)
            printf("    (IDE) %s Drive %dMB\n",
                   ide_type_names[ide_devices[i].type],
                   ide_devices[i].size / 1024 / 1024);
}
