#include <pci.h>
#include <print.h>

#include <hdd.h>

/* Informacion general sobre PCI:
 * http://wiki.osdev.org/PCI
 */

// la posta esta aca:
// https://github.com/klange/toaruos/blob/master/kernel/include/pci_list.h
static struct class_info device_info[] = {
    {0x01, 0x01, "IDE Controller"},
    {0x01, 0x05, "ATA Controller"},
    {0x01, 0x06, "Serial ATA"},
    {0x02, 0x00, "Ethernet Controller"},
    {0x03, 0x00, "VGA-Compatible Controller"},
    {0x06, 0x00, "Host Bridge"},
    {0x06, 0x01, "ISA Bridge"},
    {0x06, 0x80, "Other Host Device"},
    {0x0C, 0x03, "USB Controller"},
    {0xFF, 0x00, NULL},
};

#define DEVICE_CLASSES (sizeof(device_info) / sizeof(struct class_info))

struct pci_device pci_devices[10];
uint8_t next_pci_device = 0;

static inline uint16_t get_device(uint16_t bus, uint16_t slot, uint16_t func) {
    return ((bus << 8) | (slot << 3) | func);
}

uint32_t pci_read_config(uint16_t device, uint16_t field_and_size) {
    // Armo el pedido para informacion del dispositivo
    uint8_t field = field_and_size >> 8;
    uint8_t size = field_and_size & 0x00ff;

    uint32_t address =
        ((uint32_t)((device << 8) | (field & 0xfc))) | 0x80000000;

    // Pido esa entrada de la informacion del dispositivo
    outl(0xCF8, address);

    if (size == 4) {
        uint32_t t = inl(0xCFC);
        return t;
    } else if (size == 2) {
        uint16_t t = inw(0xCFC + (field & 2));
        return t;
    } else if (size == 1) {
        uint8_t t = inb(0xCFC + (field & 3));
        return t;
    }
    return 0xFFFF;
}

static char* lookup_device(uint8_t class, uint8_t subclass) {
    for (uint64_t i = 0; i < DEVICE_CLASSES; i++) {
        if (device_info[i].class == class &&
            device_info[i].subclass == subclass)
            return device_info[i].info;
    }
    return device_info[DEVICE_CLASSES - 1].info;
}

static void check_slot(uint8_t bus, uint8_t slot) {
    uint16_t device = get_device(bus, slot, 0);

    if (pci_read_config(device, VENDOR_ID) == 0xFFFF)
        return;  // No existe el dispositivo
    check_function(device);
    uint8_t header_type = pci_read_config(device, HEADER_TYPE);

    if ((header_type & 0x80) != 0) {
        // Es un dispositivo multifuncion,
        // entonces chequeamos todas las funciones
        for (uint8_t function = 1; function < 8; function++) {
            uint16_t device = get_device(bus, slot, function);

            if (pci_read_config(device, VENDOR_ID) != 0xFFFF)
                check_function(device);
        }
    }
}

static void check_bus(uint8_t bus) {
    for (uint8_t slot = 0; slot < 32; slot++)
        check_slot(bus, slot);
}

void check_function(uint16_t device) {
    uint8_t class = pci_read_config(device, CLASS);
    uint8_t subclass = pci_read_config(device, SUBCLASS);
    uint16_t device_id = pci_read_config(device, DEVICE_ID);
    uint16_t vendor_id = pci_read_config(device, VENDOR_ID);
    uint8_t prog_if = pci_read_config(device, PROG_IF);
    uint8_t irq = pci_read_config(device, PROG_IF);

    pci_devices[next_pci_device++] = (struct pci_device){
        device_id, vendor_id, class, subclass, prog_if, irq};

    // Imprimo toda la inforamcion sobre el dispositivo
    printf("  (PCI DEVICE) (%d,%d,%d)\t", device >> 8, (device >> 3) & 31,
           device & 3);

    char* info = lookup_device(class, subclass);
    if (info)
        printf(info);
    else
        printf("0x%x 0x%x", class, subclass);

    printf(" - [0x%x 0x%x]\n", vendor_id, device_id);
    // Aca termino de imprimir la informacion sobre el dispositivo
    // Y procedo a registrarlo como corresponda

    if (class == 0x06 && subclass == 0x04)
        check_bus(pci_read_config(device, SECONDARY_BUS));
    if (class == 0x01 && subclass == 0x01) {
        ide_init(pci_read_config(device, BAR0), pci_read_config(device, BAR1),
                 pci_read_config(device, BAR2), pci_read_config(device, BAR3),
                 pci_read_config(device, BAR4));
    }
}

void pci_check_buses(void) {
    uint8_t header_type = pci_read_config(get_device(0, 0, 0), HEADER_TYPE);
    if ((header_type & 0x80) == 0) {
        // Un solo PCI host controller
        check_bus(0);
    } else {
        // Multiples PCI host controllers. Tengo que chequear las funciones
        // (0, 0, f) . Si esta up, quiere decir que hay un host controller ahi
        for (uint8_t function = 0; function < 8; function++) {
            if (pci_read_config(get_device(0, 0, function), VENDOR_ID) !=
                0xFFFF)
                break;
            check_bus(function);
        }
    }
}
