#include <lapic.h>
#include <utils.h>

uint32_t local_apic_address = DEFAULT_LAPIC_ADDR;

#define LOCAL_APIC VOLATILE_DWORD_ARRAY(local_apic_address)

void lapic_set_address(uint32_t lapic_address) {
    local_apic_address = lapic_address;
}

void lapic_turn_on(void) {
    // Poner el valor de la entrada de la IDT para el Spurious Vector Interrupt
    // Para prender el LAPIC, hay que poner el bit 8 en 1
    LOCAL_APIC[LAPIC_SPVEC_REG] = (1 << 8) | INT_NUM_SPURIOUS;
}

// Enviar un IPI a un procesador dado su numero de local APIC.
// No confirma recepcion de la misma.
void lapic_send_ipi(const intr_command_register *ipi) {
    // clear apic errors
    // LOCAL_APIC[LAPIC_ERR_REG] = 0x0;
    const uint32_t *opts = (const uint32_t *)ipi;
    // Copiar opciones al mensaje que vamos a utilizar.
    // Se tiene que hacer de a 32 bits alineado a 16 bytes.
    // Por eso el ICR está roto en dos pedazos.
    LOCAL_APIC[LAPIC_ICR_DWORD1] = opts[1];
    // De acuerdo a Intel 3A 10.6.1, escribir la parte baja de este registro
    // hace que se envíe la interrupción. Lo escribimos al final por eso.
    LOCAL_APIC[LAPIC_ICR_DWORD0] = opts[0];
}

void lapic_wait_ipi(void) {
    // Esperar a la recepcion de la IPI.
    // El bit 12 es el bit de command completed. Cuando termine, nos vamos
    while (LOCAL_APIC[LAPIC_ICR_DWORD0] & (1 << 12));
}

// Limpia el registro de errores del LAPIC
void lapic_clear_errors(void) {
    LOCAL_APIC[LAPIC_ERR_REG] = 0x0;
}

uint8_t lapic_get_id(void) {
    // Son los 8 bits mas significativos del registro de ID
    return (LOCAL_APIC[LAPIC_ID_REG] >> 24) & 0xFF;
}

bool lapic_is_82489(void) {
    // De acuerdo al manual Intel 3A, si es un 82489DX se puede determinar
    // usando el bit 4 del byte del registro de version
    return !(LOCAL_APIC[LAPIC_VER_REG] & (1 << 4));
}

// Inicializar IPI
void lapic_init_ipi(intr_command_register *ipi, uint8_t mode, uint8_t vector,
                    uint8_t destination, uint8_t shorthand) {
    // Ponemos todo en cero para asegurar que los reserved bits estén en zero.
    *ipi = (intr_command_register){0};
    ipi->delivery_mode = mode;
    ipi->vector = vector;
    ipi->destination_field = destination;
    ipi->shorthand = shorthand;
    // Para distinguir INIT e INIT DeAssert se usan otros flags
    // Para eso utilizamos un modo separado de envio.
    if (mode == INIT_DEASSERT) {
        ipi->delivery_mode = INIT;
        ipi->shorthand = ALL_INCLUDING_SELF;
        ipi->level = 0;
        ipi->trigger_mode = 1;
    } else {
        ipi->level = 1;
        ipi->trigger_mode = 0;
    }
}
